import 'dart:convert';
import 'dart:io';

import 'package:flutter/foundation.dart';
import 'package:flutter_dotenv/flutter_dotenv.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/services.dart';

import '../../cloud_models/event.dart';
import '../../cloud_models/company_event.dart';
import '../../cloud_models/threat_weights.dart';
import '../../cloud_models/user.dart';
import '../../cloud_models/short_user.dart';
import '../../cloud_models/recommendation.dart';
import '../../cloud_models/security_defender_user.dart';
import '../../cloud_models/security_defenders_organization.dart';
import '../../cloud_models/threats.dart';
import '../../service/cloud_service/cloud_exception.dart';

class CloudService {

  late String uri;

  late HttpClient client;

  static Future<String?> _createUri() async {
    if(kReleaseMode){
      await dotenv.load(fileName: 'packages/cloud_adapter/asset/env.production');
    } else {
      await dotenv.load(fileName: 'packages/cloud_adapter/asset/env.development');
    }
    return dotenv.env['API_URI'];
  }

  static Future<HttpClient> _createClient() async {
    final certificatePath = dotenv.env['CERT_PATH']!;
    final certificateBytes = await rootBundle.load(certificatePath);
    final certificate = certificateBytes.buffer.asUint8List();

    final context = SecurityContext.defaultContext
      ..useCertificateChainBytes(
          certificate,
          password: dotenv.env['CERT_PASS']
      )
      ..usePrivateKeyBytes(
        certificate,
        password: dotenv.env['CERT_PASS']
      );

    var cl = HttpClient(context: context);
    cl.badCertificateCallback = (
        X509Certificate cert,
        String host,
        int port) {
          return true;
    };
    return cl;
  }
  CloudService._create(){
    WidgetsFlutterBinding.ensureInitialized();
  }

  static Future<CloudService> create() async {
    var c = CloudService._create();
    c.uri = await CloudService._createUri() ?? "";
    c.client = await CloudService._createClient();
    return c;
  }

  CloudService(){
    WidgetsFlutterBinding.ensureInitialized();
  }

  /// ****************
  /// EVENT OPERATIONS
  /// ****************

  // CREATE EVENT
  Future<void> createEvent(String username, Event event) async {
    try {
      print('CREATE USER EVENT');
      final String userUri = '/store/user/$username/event';
      Uri url = Uri.parse(uri + userUri);
      print(url);
      await client.postUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type', 'application/json');
            request.headers.set('accept', '*/*');
            request.write(jsonEncode(event.toJson()));
            return request.close();
          })
          .then((HttpClientResponse response) {
            if (response.statusCode == 200 || response.statusCode == 201) {
              print('EVENT CREATED');
            } else {
              print("SOMETHING WENT WRONG: " + response.statusCode.toString());
              throw Exception;
            }
          });
    } catch(e) {
      print('SOME EXCEPTION OCCURRED');
      throw Exception(e.toString());
    }
  }

  // UPDATE EVENT
  Future<void> updateEvent(String username, String eventId, Event event) async {
    try {
      print('UPDATE USER EVENT');
      final String userUri = '/store/user/$username/event/$eventId';
      Uri url = Uri.parse(uri + userUri);
      print(url);
      await client.putUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type', 'application/json');
            request.headers.set('accept', 'application/json');
            request.write(jsonEncode(event.toJson()));
            return request.close();
          })
          .then((HttpClientResponse response) {
          if (response.statusCode == 200) {
            print('USER EVENT UPDATED');
          } else {
            print("USER EVENT DID NOT UPDATE CORRECTLY: " + response.statusCode.toString());
            throw Exception;
          }
      });
    } catch (e) {
      throw CloudException("Exception updating the event from the cloud");
    }
  }

  //GET TLP WHITE EVENTS
  Future<List<Event>> getTLPWhiteEvents() async {
    List<Event> allEvents = <Event>[];
    try {
      print('TLP WHITE EVENTS');
      final String eventUri = '/store/event';
      Uri url = Uri.parse(uri + eventUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type', 'application/json');
            request.headers.set('accept', 'application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
          if (response.statusCode == 200) {
            print('RESPONSE OK');
            var result = StringBuffer();
            await for (var contents in response.transform(Utf8Decoder())) {
              result.write(contents);
            }
            List<dynamic> object = jsonDecode(result.toString());
            allEvents = object.map((e) => Event.fromJson(e)).toList();
        } else {
          print("SOMETHING WENT WRONG: " + response.statusCode.toString());
          throw Exception;
        }
      });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED');
      throw Exception(e.toString());
    }
    return allEvents;
  }

  //GET TLP WHITE EVENTS
  //FILTERED BY DATE
  Future<List<Event>> getTLPWhiteEventsDateFilter(String timestamp) async {
    List<Event> allEvents = [];
    try {
      print('TIMESTAMP FILTERED TLP WHITE EVENTS');
      final String eventUri = '/store/event';
      Uri url = Uri.parse(uri + eventUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type', 'application/json');
            request.headers.set('accept', 'application/json');
            request.headers.set('modified_since',timestamp.toString());
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if (response.statusCode == 200) {
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              List<dynamic> object = jsonDecode(result.toString());
              allEvents = object.map((e) => Event.fromJson(e)).toList();
            } else {
              print("SOMETHING WENT WRONG: " + response.statusCode.toString());
              throw Exception;
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED');
      print(e);
      throw Exception;
    }
    return allEvents;
  }


  Future<List<Event>> getEventsByTypesAndOwner(List<String> types, String owner) async {
    List<Event> allEvents = [];
    try {
      print('GET EVENTS BY TYPES AND OWNER');
      final String eventUri = '/store/events';
      Uri url = Uri.parse(uri + eventUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type', 'application/json');
            request.headers.set('accept', 'application/json');
            request.headers.set('type',types.toString().replaceAll('[','').replaceAll(']',''));
            request.headers.set('owner', owner);
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if (response.statusCode == 200) {
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              List<dynamic> object = jsonDecode(result.toString());
              allEvents = object.map((e) => Event.fromJson(e)).toList();
            } else {
              print("SOMETHING WENT WRONG: " + response.statusCode.toString());
              throw Exception;
            }
          });
    } catch(e) {
      print('SOME EXCEPTION OCCURRED');
      print(e);
      throw Exception;
    }
    return allEvents;
  }

  Future<List<Event>> getEventsByTypesAndOwnerAndAfter(List<String> types, String owner, String timestamp) async {
    List<Event> allEvents = [];
    try {
      print('GET EVENTS BY TYPES AND OWNER AND AFTER TIMESTAMP');
      final String eventUri = '/store/events/after';
      Uri url = Uri.parse(uri + eventUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type', 'application/json');
            request.headers.set('accept', 'application/json');
            request.headers.set('type',types.toString().replaceAll('[','').replaceAll(']',''));
            request.headers.set('owner', owner);
            request.headers.set('modified_since', timestamp.toString());
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if (response.statusCode == 200) {
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              List<dynamic> object = jsonDecode(result.toString());
              allEvents = object.map((e) => Event.fromJson(e)).toList();
            } else {
              print("SOMETHING WENT WRONG: " + response.statusCode.toString());
              throw Exception;
            }
          });
    } catch(e) {
      print('SOME EXCEPTION OCCURRED');
      print(e);
      throw Exception;
    }
    return allEvents;
  }

  //GET LIST OF USER EVENTS
  Future<List<String>> getUserEvents(String userId) async {
    List<String> allEvents = [];
    try {
      print('GET USER EVENT LIST');
      final String eventUri = '/store/user/$userId/event';
      Uri url = Uri.parse(uri + eventUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type', 'application/json');
            request.headers.set('accept', 'application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if (response.statusCode == 200) {
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              List<dynamic> object = jsonDecode(result.toString());
              allEvents = object.cast<String>();
            } else {
              print("SOMETHING WENT WRONG: " + response.statusCode.toString());
              throw Exception;
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED');
      print(e);
      throw Exception;
    }
    return allEvents;
  }

  //GET LIST OF USER EVENTS
  //FILTERED BY DATE
  Future<List<String>> getUserEventsDateFilter(String userId, String timestamp) async {
    List<String> allEvents = [];
    try {
      print('TIMESTAMP FILTERED GET USER EVENT LIST');
      final String eventUri = '/store/user/$userId/event';
      Uri url = Uri.parse(uri + eventUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type', 'application/json');
            request.headers.set('accept', 'application/json');
            request.headers.set('modified_since', timestamp.toString());
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if (response.statusCode == 200) {
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              List<dynamic> object = jsonDecode(result.toString());
              allEvents = object.map((e) => e.toString()).toList();
            } else {
              print("SOMETHING WENT WRONG: " + response.statusCode.toString());
              throw Exception;
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED');
      print(e);
      throw Exception;
    }
    return allEvents;
  }

  //GET USER EVENT
  Future<Event> getSingleUserEvent(String userId, String eventId) async {
    Event event = Event(id_event: eventId);
    try {
      print('GET SINGLE USER EVENT');
      final String eventUri = '/store/user/$userId/event/$eventId';
      Uri url = Uri.parse(uri + eventUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type', 'application/json');
            request.headers.set('accept', 'application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if (response.statusCode == 200) {
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              var object = jsonDecode(result.toString());
              event = Event.fromJson(object);
            } else {
              print("SOMETHING WENT WRONG: " + response.statusCode.toString());
              throw Exception;
            }
          });
    } catch (e) {
      print(e);
      throw CloudException("Exception getting the event from the cloud");
    }
    return event;
  }

  //DELETE USER EVENT
  Future<void> deleteEvent(String username, String eventId) async {
    try {
      print('DELETE EVENT');
      final String eventUri = '/store/user/$username/event/$eventId';
      Uri url = Uri.parse(uri + eventUri);
      print(url);
      await client.deleteUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('accept', '*/*');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if (response.statusCode != 200) {
              print("SOMETHING WENT WRONG: " + response.statusCode.toString());
              throw CloudException("Exception removing the event. Status code: " +
                  response.statusCode.toString());
            }
          });
    } catch (e) {
      print(e);
      throw CloudException("Exception removing the event from the cloud");
    }
  }

  /// ***************
  /// USER OPERATIONS
  /// ***************

  //CHECK IF A USER EXIST
  Future<bool> userExists(String username) async {
    bool result = false;
    try {
      print('CHECK USER');
      final String userUri = '/store/user/$username';
      Uri url = Uri.parse(uri + userUri);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type','application/json');
            request.headers.set('accept','application/json');
            return request.close();
          })
          .then((HttpClientResponse response) {
            if(response.statusCode == 200){
              print('USER EXISTS');
              result = true;
            } else {
              if (response.statusCode == 404) {
                print("USER NOT EXIST");
              } else {
                print(response.statusCode.toString());
                throw Exception;
              }
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED' + e.toString());
    }
    return result;
  }

  //GET USER
  Future<User> getUser(String username) async {
    User user = User(idUser: '');
    try {
      print('GET USER');
      final String userUri = '/store/user/$username';
      Uri url = Uri.parse(uri + userUri);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type','application/json');
            request.headers.set('accept','application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if(response.statusCode == 200){
              print('USER EXISTS');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              var object = jsonDecode(result.toString());
              user = User.fromJson(object);
            } else {
              if (response.statusCode == 404) {
                print("USER NOT EXIST");
              } else {
                print(response.statusCode.toString());
                throw Exception;
              }
            }
          });
    } catch (e) {
      print(e);
      throw CloudException('This should not happen');
    }
    return user;
  }

  //CREATE A NEW USER
  Future<void> createUser(String username,
      String? email,
      String? access,
      String? expires,
      String? name,
      String? publicKey) async {
    try {
      print('CREATE USER');
      final String userUri = '/store/user';
      Uri url = Uri.parse(uri + userUri);
      var body = {'id_user': username};
      if (email != null) {
        body['email'] = email;
      }
      if (access != null) {
        body['access'] = access;
      }
      if (expires != null) {
        body['expires'] = expires;
      }
      if (name != null) {
        body['name'] = name;
      }
      if (publicKey != null) {
        body['public_key'] = publicKey;
      }

      await client.postUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type', 'application/json');
            request.headers.set('accept', '*/*');
            request.write(jsonEncode(body));
            return request.close();
          })
          .then((HttpClientResponse response) {
            if (response.statusCode == 200) {
              print("USER "+ username + " CREATED");
            } else {
              print("FAILED: STATUS CODE " + response.statusCode.toString());
              throw Exception;
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED' + e.toString());
      throw Exception;
    }
  }

  // UPDATE USER
  Future<void> updateUser(String username, String name) async {
    try {
      print('UPDATE USER');
      var body = {'id_user': username};
      if (name.isNotEmpty) {
        body['name'] = name;
      }
      final String userUri = '/store/user/$username';
      Uri url = Uri.parse(uri + userUri);
      await client.putUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type', 'application/json');
            request.headers.set('accept', 'application/json');
            request.write(jsonEncode(body));
            return request.close();
          })
          .then((HttpClientResponse response) {
            if (response.statusCode == 200) {
              print('USER UPDATED');
            } else {
              print("FAILED: UPDATING USER STATUS CODE " + response.statusCode.toString());
              throw Exception;
            }
          });
    } catch (e) {
      throw CloudException("Exception updating user in the cloud");
    }
  }

  //DELETE USER INFORMATION
  Future<void> deleteUser(String idUser1) async {
    try {
      print('DELETE USER INFORMATION');
      final String eventUri = '/store/user/$idUser1';
      Uri url = Uri.parse(uri + eventUri);
      print(url);

      await client.deleteUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('accept', '*/*');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if (response.statusCode != 200) {
              print("SOMETHING WENT WRONG: " + response.statusCode.toString());
              throw CloudException("Error deleting USER information from cloud. Status code: " +
              response.statusCode.toString());
            }
          });
    } catch (e) {
      print(e);
      throw CloudException("Error deleting user information from cloud.");
    }
  }

  //GET LIST OF USERS
  Future<List<User>> getUsers() async {
    List<User> usersList = [];
    try {
      print('GET USER LIST');
      final String userUri = '/store/user';
      Uri url = Uri.parse(uri + userUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type','application/json');
            request.headers.set('accept','application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if(response.statusCode == 200){
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              var object = jsonDecode(result.toString());
              usersList = object.map((e) => User.fromJson(e)).toList();
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED' + e.toString());
      throw Exception;
    }
    return usersList;
  }

  //GET USER GEIGER SCORE
  Future<Event> getUserGeigerScore(String userId) async {
    Event event = Event(id_event: '');
    try {
      print('GET USER GEIGER SCORE');
      final String userUri = '/store/user/$userId/geigerscore';
      Uri url = Uri.parse(uri + userUri);
      print(url);

      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type','application/json');
            request.headers.set('accept','application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if(response.statusCode == 200){
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              var object = jsonDecode(result.toString());
              event = Event.fromJson(object);
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED' + e.toString());
      throw Exception;
    }
    return event;
  }

  // GET USER EVENT KEY
  Future<Event> getUserEventKey(String userId, String path) async {
    Event event = Event(id_event: '');
    try {
      print('GET USER EVENT KEY');
      final String userUri = '/store/user/$userId/$path/eventkey';
      Uri url = Uri.parse(uri + userUri);
      print(url);

      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type','application/json');
            request.headers.set('accept','application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if(response.statusCode == 200){
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              var object = jsonDecode(result.toString());
              event = Event.fromJson(object);
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED' + e.toString());
      throw Exception;
    }
    return event;
  }

  //GET USER DEVICE INFO
  Future<Event> getUserDeviceInfo(String userId, String deviceId) async {
    Event event = Event(id_event: '');
    try {
      print('GET USER DEVICE INFO');
      final String userUri = '/store/user/$userId/device/$deviceId';
      Uri url = Uri.parse(uri + userUri);
      print(url);

      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type','application/json');
            request.headers.set('accept','application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if(response.statusCode == 200){
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              var object = jsonDecode(result.toString());
              event = Event.fromJson(object);
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED' + e.toString());
      throw Exception;
    }
    return event;
  }
  /// USER MERGED DATA

  // MERGE INFORMATION
  Future<void> createMerge(String idUser1, String idUser2, String agreement,
      [String? type]) async {
    try {
      print('CREATE MERGE');
      String userUri =
          '/store/user/$idUser1/merge/$idUser2?agreement=$agreement';
      if (type != null) {
        userUri = userUri + '&type=$type';
      }
      Uri url = Uri.parse(uri + userUri);
      print(url.toString());
      await client.postUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('accept','*/*');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if(response.statusCode == 200){
              print('USER MERGED');
            } else {
              print("FAILED FUNCTION createMerge STATUS CODE: " +
              response.statusCode.toString());
              throw Exception;
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED' + e.toString());
      throw Exception;
    }
  }

  // GET LIST OF USER MERGED ACCOUNTS
  Future<List<String>> getMergedAccounts(String idUser) async {
    List<String> allMerged = [];
    try {
      print('GET MERGED LIST');
      final String mergedUri = '/store/user/$idUser/merge';
      Uri url = Uri.parse(uri + mergedUri);
      print(url);

      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type','application/json');
            request.headers.set('accept','application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if(response.statusCode == 200){
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              List<dynamic> object = jsonDecode(result.toString());
              allMerged = object.map((e) => e.toString()).toList();
            }
          });
    } catch (e) {
      print(e);
      throw CloudException("Failure getting merged account from FUNCTION getMergedAccounts: "+e.toString());
    }
    return allMerged;
  }

  // GET INFO OF ONE MERGED ACCOUNT
  Future<ShortUser> getMergedInfo(String idUser1, String idUser2) async {
    ShortUser sUser = ShortUser(idUser: idUser1);
    try {
      print('GET MERGED INFORMATION');
      final String mergedUri = '/store/user/$idUser1/merge/$idUser2';
      Uri url = Uri.parse(uri + mergedUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type','application/json');
            request.headers.set('accept','application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if(response.statusCode == 200){
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              var object = jsonDecode(result.toString());
              sUser = ShortUser.fromJson(object);
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED' + e.toString());
      throw Exception;
    }
    return sUser;
  }

  //DELETE MERGED INFORMATION
  Future<void> deleteMerged(String idUser1, String idUser2, String type) async {
    try {
      print('DELETE MERGED INFORMATION');
      final String eventUri = '/store/user/$idUser1/merge/$idUser2/$type';
      Uri url = Uri.parse(uri + eventUri);
      print(url);
      await client.deleteUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('accept', '*/*');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if (response.statusCode != 200) {
              print("SOMETHING WENT WRONG: " + response.statusCode.toString());
              throw CloudException("Error deleting agreement from cloud. Status code: " +
              response.statusCode.toString());
            } else if(response.statusCode == 200) {
              print('MERGE DELETED OK');
            }
          });
    } catch (e) {
      print(e);
      throw CloudException("Error deleting agreement from cloud.");
    }
  }

  //GET COMPANY EVENTS
  Future<List<String>> getCompanyEvents(String companyId) async {
    List<String> allEvents = [];
    try {
      print('GET COMPANY EVENTS');
      final String eventUri = '/store/company/$companyId/event';
      Uri url = Uri.parse(uri + eventUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type','application/json');
            request.headers.set('accept','application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if(response.statusCode == 200){
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              List<dynamic> object = jsonDecode(result.toString());
              allEvents = object.map((e) => e.toString()).toList();
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED' + e.toString());
      throw Exception;
    }
    return allEvents;
  }

  //GET SINGLE COMPANY EVENT
  Future<CompanyEvent> getSingleCompanyEvent(String companyId, String eventId) async {
    CompanyEvent cEvent = CompanyEvent();
    try {
      print('GET SINGLE COMPANY EVENT');
      final String eventUri = '/store/company/$companyId/event/$eventId';
      Uri url = Uri.parse(uri + eventUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type','application/json');
            request.headers.set('accept','application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if(response.statusCode == 200){
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              var object = jsonDecode(result.toString());
              cEvent = CompanyEvent.fromJson(object);
            }
          });
    } catch (e) {
      print(e);
      throw CloudException("Exception getting the event from the cloud");
    }
    return cEvent;
  }

  /// *************************
  /// THREAT WEIGHTS OPERATIONS
  /// *************************

  //GET THREAT WEIGHTS
  Future<List<ThreatWeights>> getThreatWeights() async {
    List<ThreatWeights> threatWeights = [];
    try {
      print('GET THREAT WEIGHTS');
      final String threatUri = '/threatweights';
      Uri url = Uri.parse(uri + threatUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type', 'application/json');
            request.headers.set('accept', 'application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if (response.statusCode == 200) {
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              List<dynamic> object = jsonDecode(result.toString());
              threatWeights = object.map((e) => ThreatWeights.fromJson(e)).toList();
            } else {
              print("SOMETHING WENT WRONG: " + response.statusCode.toString());
              throw Exception;
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED' + e.toString());
      throw Exception;
    }
    return threatWeights;
  }

  /// *************************
  /// GEIGER THREATS OPERATIONS
  /// *************************
  Future<List<Threats>> getThreats() async {
    List<Threats> threats = [];
    try {
      print('GET THREATS');
      final String threatUri = '/threats';
      Uri url = Uri.parse(uri + threatUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type', 'application/json');
            request.headers.set('accept', 'application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if (response.statusCode == 200) {
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              List<dynamic> object = jsonDecode(result.toString());
              threats = object.map((e) => Threats.fromJson(e)).toList();
            } else {
              print("SOMETHING WENT WRONG: " + response.statusCode.toString());
              throw Exception;
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED' + e.toString());
      throw Exception;
    }
    return threats;
  }

  /// *************************
  /// RECOMMENDATIONS OPERATIONS
  /// *************************

  //GET RECOMMENDATIONS
  Future<List<Recommendation>> getRecommendations() async {
    List<Recommendation> recommendations = [];
    try {
      print('GET RECOMMENDATIONS');
      final String threatUri = '/recommendation';
      Uri url = Uri.parse(uri + threatUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type', 'application/json');
            request.headers.set('accept', 'application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if (response.statusCode == 200) {
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              List<dynamic> object = jsonDecode(result.toString());
              recommendations = object.map((e) => Recommendation.fromJson(e)).toList();
            } else {
              print("SOMETHING WENT WRONG: " + response.statusCode.toString());
              throw Exception;
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED' + e.toString());
      throw Exception;
    }
    return recommendations;
  }

  /// *************************
  /// SECURITY DEFENDERS INFO OPERATIONS
  /// *************************
  Future<List<SecurityDefendersOrganization>>
      getSecurityDefendersOrganizations() async {
    List<SecurityDefendersOrganization> securityList = [];
    try {
      print('GET SECURITY DEFENDERS ORGANIZATIONS');
      final String secUri = '/security-defenders-organizations';
      Uri url = Uri.parse(uri + secUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type', 'application/json');
            request.headers.set('accept', 'application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if (response.statusCode == 200) {
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              List<dynamic> object = jsonDecode(result.toString());
              securityList = object.map((e) => SecurityDefendersOrganization.fromJson(e)).toList();
            } else {
              print("SOMETHING WENT WRONG: " + response.statusCode.toString());
              throw Exception;
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED' + e.toString());
      throw Exception;
    }
    return securityList;
  }

  Future<List<String>> getSecurityDefendersOrganizationUsers(String idSecOrg) async {
    List<String> securityList = [];
    try {
      print('GET USER EVENT LIST');
      final String eventUri = '/store/user/$idSecOrg/event';
      Uri url = Uri.parse(uri + eventUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type','application/json');
            request.headers.set('accept','application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if(response.statusCode == 200){
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              List<dynamic> object = jsonDecode(result.toString());
              securityList = object.map((e) => e.toString()).toList();
            }
          });
    } catch (e) {
      print('SOME EXCEPTION OCCURRED' + e.toString());
      throw Exception;
    }
    return securityList;
  }

  Future<SecurityDefenderUser> getSecurityDefenderUser(String idSecOrg, String idSerOrgUser) async {
    SecurityDefenderUser secUser = SecurityDefenderUser(id_user: idSerOrgUser);
    try {
      print('GET SINGLE USER EVENT');
      final String eventUri = '/store/user/$idSecOrg/event/$idSerOrgUser';
      Uri url = Uri.parse(uri + eventUri);
      print(url);
      await client.getUrl(url)
          .then((HttpClientRequest request) {
            request.headers.set('content-type','application/json');
            request.headers.set('accept','application/json');
            return request.close();
          })
          .then((HttpClientResponse response) async {
            if(response.statusCode == 200){
              print('RESPONSE OK');
              var result = StringBuffer();
              await for (var contents in response.transform(Utf8Decoder())) {
                result.write(contents);
              }
              var object = jsonDecode(result.toString());
              secUser = SecurityDefenderUser.fromJson(object);
            }
          });
    } catch (e) {
      print(e);
      throw CloudException("Exception getting the event from the cloud");
    }
    return secUser;
  }

  /// *******************
  /// CLOUD UTILS METHODS
  /// *******************
  List<Event> parseEvent(String responseBody) {
    final parsed = jsonDecode(responseBody).cast<Map<String, dynamic>>();
    return parsed.map<Event>((e) => Event.fromJson(e)).toList();
  }

  Future<String> generateUUID() async {
    String uuid = '';
    Uri url = Uri.parse(uri + '/uuid');
    await client.getUrl(url)
        .then((HttpClientRequest request) {
          request.headers.set('accept','text/plain');
          return request.close();
        })
        .then((HttpClientResponse response) async {
          if(response.statusCode == 200){
            print('RESPONSE OK');
            var result = StringBuffer();
            await for (var contents in response.transform(Utf8Decoder())) {
              result.write(contents);
            }
            uuid = result.toString();
          } else {
            throw CloudException("FAILURE GETTING A UUID");
          }
        });
    return uuid;
  }
}
